@@ -24,14 +24,47 @@ from .constants import (
 from .container import WidgetContainerListContentsMixin, WidgetContainerMixin
 from .filler import calculate_top_bottom_filler
 from .padding import calculate_left_right_padding
-from .widget import Widget
+from .widget import Widget, WidgetError, WidgetWarning
 
 if typing.TYPE_CHECKING:
+    from collections.abc import Iterator
+
     from typing_extensions import Literal
 
 
-class OverlayError(Exception):
-    pass
+class OverlayError(WidgetError):
+    """Overlay specific errors."""
+
+
+class OverlayWarning(WidgetWarning):
+    """Overlay specific warnings."""
+
+
+def _check_widget_subclass(widget: Widget) -> None:
+    if not isinstance(widget, Widget):
+        obj_class_path = f"{widget.__class__.__module__}.{widget.__class__.__name__}"
+        warnings.warn(
+            f"{obj_class_path} is not subclass of Widget",
+            DeprecationWarning,
+            stacklevel=3,
+        )
+
+
+class OverlayOptions(typing.NamedTuple):
+    align: Align
+    align_amount: int | None
+    width_type: WHSettings
+    width_amount: int | None
+    min_width: int | None
+    left: int
+    right: int
+    valign_type: VAlign
+    valign_amount: int | None
+    height_type: WHSettings
+    height_amount: int | None
+    min_height: int | None
+    top: int
+    bottom: int
 
 
 class Overlay(Widget, WidgetContainerMixin, WidgetContainerListContentsMixin):
@@ -40,23 +73,22 @@ class Overlay(Widget, WidgetContainerMixin, WidgetContainerListContentsMixin):
     """
 
     _selectable = True
-    _sizing = frozenset([Sizing.BOX])
 
-    _DEFAULT_BOTTOM_OPTIONS = (
-        Align.LEFT,
-        None,
-        WHSettings.RELATIVE,
-        100,
-        None,
-        0,
-        0,
-        VAlign.TOP,
-        None,
-        WHSettings.RELATIVE,
-        100,
-        None,
-        0,
-        0,
+    _DEFAULT_BOTTOM_OPTIONS = OverlayOptions(
+        align=Align.LEFT,
+        align_amount=None,
+        width_type=WHSettings.RELATIVE,
+        width_amount=100,
+        min_width=None,
+        left=0,
+        right=0,
+        valign_type=VAlign.TOP,
+        valign_amount=None,
+        height_type=WHSettings.RELATIVE,
+        height_amount=100,
+        min_height=None,
+        top=0,
+        bottom=0,
     )
 
     def __init__(
@@ -109,11 +141,9 @@ class Overlay(Widget, WidgetContainerMixin, WidgetContainerListContentsMixin):
               integer number of rows high
             (``'relative'``, *percentage of total height*)
               make *top_w* height related to container height
-        :param min_width: the minimum number of columns for *top_w* when width
-            is not fixed
+        :param min_width: the minimum number of columns for *top_w* when width is not fixed
         :type min_width: int
-        :param min_height: minimum number of rows for *top_w* when height
-            is not fixed
+        :param min_height: minimum number of rows for *top_w* when height is not fixed
         :type min_height: int
         :param left: a fixed number of columns to add on the left
         :type left: int
@@ -135,6 +165,166 @@ class Overlay(Widget, WidgetContainerMixin, WidgetContainerListContentsMixin):
 
         self.set_overlay_parameters(align, width, valign, height, min_width, min_height, left, right, top, bottom)
 
+        _check_widget_subclass(top_w)
+        _check_widget_subclass(bottom_w)
+
+    def sizing(self) -> frozenset[Sizing]:
+        """Actual widget sizing.
+
+        :returns: Sizing information depends on the top widget sizing and sizing parameters.
+        :rtype: frozenset[Sizing]
+
+        Rules:
+        * BOX sizing is always supported provided by the bottom widget
+        * FLOW sizing is supported if top widget has:
+        * * PACK height type and FLOW supported by the TOP widget
+        * * BOX supported by TOP widget AND height amount AND height type GIVEN of min_height
+        * FIXED sizing is supported if top widget has:
+        * * PACK width type and FIXED supported by the TOP widget
+        * * width amount and GIVEN width or min_width AND:
+        * * * FLOW supported by the TOP widget AND PACK height type
+        * * * BOX supported by the TOP widget AND height_amount and GIVEN height or min height
+        """
+        sizing = {Sizing.BOX}
+        top_sizing = self.top_w.sizing()
+        if self.width_type == WHSettings.PACK:
+            if Sizing.FIXED in top_sizing:
+                sizing.add(Sizing.FIXED)
+            else:
+                warnings.warn(
+                    f"Top widget {self.top_w} should support sizing {Sizing.FIXED.upper()} "
+                    f"for width type {WHSettings.PACK.upper()}",
+                    OverlayWarning,
+                    stacklevel=3,
+                )
+
+        elif self.height_type == WHSettings.PACK:
+            if Sizing.FLOW in top_sizing:
+                sizing.add(Sizing.FLOW)
+
+                if self.width_amount and (self.width_type == WHSettings.GIVEN or self.min_width):
+                    sizing.add(Sizing.FIXED)
+            else:
+                warnings.warn(
+                    f"Top widget {self.top_w} should support sizing {Sizing.FLOW.upper()} "
+                    f"for height type {self.height_type.upper()}",
+                    OverlayWarning,
+                    stacklevel=3,
+                )
+
+        elif self.height_amount and (self.height_type == WHSettings.GIVEN or self.min_height):
+            if Sizing.BOX in top_sizing:
+                sizing.add(Sizing.FLOW)
+                if self.width_amount and (self.width_type == WHSettings.GIVEN or self.min_width):
+                    sizing.add(Sizing.FIXED)
+            else:
+                warnings.warn(
+                    f"Top widget {self.top_w} should support sizing {Sizing.BOX.upper()} "
+                    f"for height type {self.height_type.upper()}",
+                    OverlayWarning,
+                    stacklevel=3,
+                )
+
+        return frozenset(sizing)
+
+    def pack(self, size: tuple[()] | tuple[int] | tuple[int, int] = (), focus: bool = False) -> tuple[int, int]:
+        if size:
+            return super().pack(size, focus)
+
+        extra_cols = (self.left or 0) + (self.right or 0)
+        extra_rows = (self.top or 0) + (self.bottom or 0)
+
+        if self.width_type == WHSettings.PACK:
+            cols, rows = self.top_w.pack((), focus)
+            return cols + extra_cols, rows + extra_rows
+
+        if not self.width_amount:
+            raise OverlayError(
+                f"Requested FIXED render for {self.top_w} with "
+                f"width_type={self.width_type.upper()}, "
+                f"width_amount={self.width_amount!r}, "
+                f"height_type={self.height_type.upper()}, "
+                f"height_amount={self.height_amount!r}"
+                f"min_width={self.min_width!r}, "
+                f"min_height={self.min_height!r}"
+            )
+
+        if self.width_type == WHSettings.GIVEN:
+            w_cols = self.width_amount
+            cols = w_cols + extra_cols
+        elif self.width_type == WHSettings.RELATIVE and self.min_width:
+            w_cols = self.min_width
+            cols = int(w_cols * 100 / self.width_amount + 0.5)
+        else:
+            raise OverlayError(
+                f"Requested FIXED render for {self.top_w} with "
+                f"width_type={self.width_type.upper()}, "
+                f"width_amount={self.width_amount!r}, "
+                f"height_type={self.height_type.upper()}, "
+                f"height_amount={self.height_amount!r}"
+                f"min_width={self.min_width!r}, "
+                f"min_height={self.min_height!r}"
+            )
+
+        if self.height_type == WHSettings.PACK:
+            return cols, self.top_w.rows((w_cols,), focus) + extra_rows
+
+        if not self.height_amount:
+            raise OverlayError(
+                f"Requested FIXED render for {self.top_w} with "
+                f"width_type={self.width_type.upper()}, "
+                f"width_amount={self.width_amount!r}, "
+                f"height_type={self.height_type.upper()}, "
+                f"height_amount={self.height_amount!r}"
+                f"min_width={self.min_width!r}, "
+                f"min_height={self.min_height!r}"
+            )
+
+        if self.height_type == WHSettings.GIVEN:
+            return cols, self.height_amount + extra_rows
+
+        if self.height_type == WHSettings.RELATIVE and self.min_height:
+            return cols, int(self.min_height * 100 / self.height_amount + 0.5)
+
+        raise OverlayError(
+            f"Requested FIXED render for {self.top_w} with "
+            f"width_type={self.width_type.upper()}, "
+            f"width_amount={self.width_amount!r}, "
+            f"height_type={self.height_type.upper()}, "
+            f"height_amount={self.height_amount!r}"
+            f"min_width={self.min_width!r}, "
+            f"min_height={self.min_height!r}"
+        )
+
+    def rows(self, size: tuple[int], focus: bool = False) -> int:
+        """Widget rows amount for FLOW sizing."""
+        extra_height = (self.top or 0) + (self.bottom or 0)
+        if self.height_type == WHSettings.GIVEN:
+            return self.height_amount + extra_height
+        if self.height_type == WHSettings.RELATIVE and self.min_height:
+            return int(self.min_height * 100 / self.height_amount + 0.5)
+
+        if self.height_type == WHSettings.PACK:
+            extra_height = (self.top or 0) + (self.bottom or 0)
+            if self.width_type == WHSettings.GIVEN and self.width_amount:
+                return self.top_w.rows((self.width_amount,), focus) + extra_height
+            if self.width_type == WHSettings.RELATIVE:
+                width = max(int(size[0] * self.width_amount / 100 + 0.5), (self.min_width or 0))
+                return self.top_w.rows((width,), focus) + extra_height
+
+        raise OverlayError(
+            f"Requested rows for {self.top_w} with size {size!r}"
+            f"width_type={self.width_type.upper()}, "
+            f"width_amount={self.width_amount!r}, "
+            f"height_type={self.height_type.upper()}, "
+            f"height_amount={self.height_amount!r}"
+            f"min_width={self.min_width!r}, "
+            f"min_height={self.min_height!r}"
+        )
+
+    def _repr_words(self) -> list[str]:
+        return [*super()._repr_words(), f"top={self.top_w!r}", f"bottom={self.bottom_w!r}"]
+
     @staticmethod
     def options(
         align_type: Literal["left", "center", "right", "relative"] | Align,
@@ -151,7 +341,7 @@ class Overlay(Widget, WidgetContainerMixin, WidgetContainerListContentsMixin):
         right: int = 0,
         top: int = 0,
         bottom: int = 0,
-    ):
+    ) -> OverlayOptions:
         """
         Return a new options tuple for use in this Overlay's .contents mapping.
 
@@ -161,17 +351,17 @@ class Overlay(Widget, WidgetContainerMixin, WidgetContainerListContentsMixin):
         See also :meth:`.set_overlay_parameters`
         """
 
-        return (
-            align_type,
+        return OverlayOptions(
+            Align(align_type),
             align_amount,
-            width_type,
+            WHSettings(width_type),
             width_amount,
             min_width,
             left,
             right,
-            valign_type,
+            VAlign(valign_type),
             valign_amount,
-            height_type,
+            WHSettings(height_type),
             height_amount,
             min_height,
             top,
@@ -181,20 +371,24 @@ class Overlay(Widget, WidgetContainerMixin, WidgetContainerListContentsMixin):
     def set_overlay_parameters(
         self,
         align: (
-            Literal["left", "center", "right"] | Align | tuple[Literal["relative", "fixed left", "fixed right"], int]
+            Literal["left", "center", "right"]
+            | Align
+            | tuple[Literal["relative", "fixed left", "fixed right", WHSettings.RELATIVE], int]
         ),
-        width: int | None,
+        width: Literal["pack", WHSettings.PACK] | int | tuple[Literal["relative", WHSettings.RELATIVE], int] | None,
         valign: (
-            Literal["top", "middle", "bottom"] | VAlign | tuple[Literal["relative", "fixed top", "fixed bottom"], int]
+            Literal["top", "middle", "bottom"]
+            | VAlign
+            | tuple[Literal["relative", "fixed top", "fixed bottom", WHSettings.RELATIVE], int]
         ),
-        height: int | None,
+        height: Literal["pack", WHSettings.PACK] | int | tuple[Literal["relative", WHSettings.RELATIVE], int] | None,
         min_width: int | None = None,
         min_height: int | None = None,
         left: int = 0,
         right: int = 0,
         top: int = 0,
         bottom: int = 0,
-    ):
+    ) -> None:
         """
         Adjust the overlay size and position parameters.
 
@@ -205,10 +399,15 @@ class Overlay(Widget, WidgetContainerMixin, WidgetContainerListContentsMixin):
         if isinstance(align, tuple):
             if align[0] == "fixed left":
                 left = align[1]
-                align = Align.LEFT
+                normalized_align = Align.LEFT
             elif align[0] == "fixed right":
                 right = align[1]
-                align = Align.RIGHT
+                normalized_align = Align.RIGHT
+            else:
+                normalized_align = align
+        else:
+            normalized_align = Align(align)
+
         if isinstance(width, tuple):
             if width[0] == "fixed left":
                 left = width[1]
@@ -216,13 +415,23 @@ class Overlay(Widget, WidgetContainerMixin, WidgetContainerListContentsMixin):
             elif width[0] == "fixed right":
                 right = width[1]
                 width = RELATIVE_100
+
         if isinstance(valign, tuple):
             if valign[0] == "fixed top":
                 top = valign[1]
-                valign = VAlign.TOP
+                normalized_valign = VAlign.TOP
             elif valign[0] == "fixed bottom":
                 bottom = valign[1]
-                valign = VAlign.BOTTOM
+                normalized_valign = VAlign.BOTTOM
+            else:
+                normalized_valign = valign
+
+        elif not isinstance(valign, (VAlign, str)):
+            raise OverlayError(f"invalid valign: {valign!r}")
+
+        else:
+            normalized_valign = VAlign(valign)
+
         if isinstance(height, tuple):
             if height[0] == "fixed bottom":
                 bottom = height[1]
@@ -236,12 +445,12 @@ class Overlay(Widget, WidgetContainerMixin, WidgetContainerListContentsMixin):
         if height is None:
             height = WHSettings.PACK
 
-        align_type, align_amount = normalize_align(align, OverlayError)
+        align_type, align_amount = normalize_align(normalized_align, OverlayError)
         width_type, width_amount = normalize_width(width, OverlayError)
-        valign_type, valign_amount = normalize_valign(valign, OverlayError)
+        valign_type, valign_amount = normalize_valign(normalized_valign, OverlayError)
         height_type, height_amount = normalize_height(height, OverlayError)
 
-        if height_type in (WHSettings.GIVEN, WHSettings.PACK):
+        if height_type in {WHSettings.GIVEN, WHSettings.PACK}:
             min_height = None
 
         # use container API to set the parameters
@@ -269,9 +478,10 @@ class Overlay(Widget, WidgetContainerMixin, WidgetContainerListContentsMixin):
         """Return selectable from top_w."""
         return self.top_w.selectable()
 
-    def keypress(self, size: tuple[int, int], key: str) -> str | None:
+    def keypress(self, size: tuple[()] | tuple[int] | tuple[int, int], key: str) -> str | None:
         """Pass keypress to top_w."""
-        return self.top_w.keypress(self.top_w_size(size, *self.calculate_padding_filler(size, True)), key)
+        real_size = self.pack(size, True)
+        return self.top_w.keypress(self.top_w_size(real_size, *self.calculate_padding_filler(real_size, True)), key)
 
     @property
     def focus(self) -> Widget:
@@ -356,13 +566,25 @@ class Overlay(Widget, WidgetContainerMixin, WidgetContainerListContentsMixin):
         :exc:`OverlayError`.
         """
 
-        class OverlayContents:
-            def __len__(inner_self):
+        # noinspection PyMethodParameters
+        class OverlayContents(typing.Sequence[typing.Tuple[Widget, OverlayOptions]]):
+            def __len__(inner_self) -> int:
                 return 2
 
             __getitem__ = self._contents__getitem__
             __setitem__ = self._contents__setitem__
 
+            def __repr__(inner_self) -> str:
+                return repr(f"<{inner_self.__class__.__name__}({[inner_self[0], inner_self[1]]})> for {self}")
+
+            def __rich_repr__(inner_self) -> Iterator[tuple[str | None, typing.Any] | typing.Any]:
+                for idx in range(len(inner_self)):
+                    yield None, inner_self[idx]
+
+            def __iter__(inner_self) -> Iterator[tuple[Widget, OverlayOptions]]:
+                for idx in range(2):
+                    yield inner_self[idx]
+
         return OverlayContents()
 
     @contents.setter
@@ -372,13 +594,14 @@ class Overlay(Widget, WidgetContainerMixin, WidgetContainerListContentsMixin):
         self.contents[0] = new_contents[0]
         self.contents[1] = new_contents[1]
 
-    def _contents__getitem__(self, index: Literal[0, 1]):
+    def _contents__getitem__(self, index: Literal[0, 1]) -> tuple[Widget, OverlayOptions]:
         if index == 0:
             return (self.bottom_w, self._DEFAULT_BOTTOM_OPTIONS)
+
         if index == 1:
             return (
                 self.top_w,
-                (
+                OverlayOptions(
                     self.align_type,
                     self.align_amount,
                     self.width_type,
@@ -397,7 +620,7 @@ class Overlay(Widget, WidgetContainerMixin, WidgetContainerListContentsMixin):
             )
         raise IndexError(f"Overlay.contents has no position {index!r}")
 
-    def _contents__setitem__(self, index: Literal[0, 1], value):
+    def _contents__setitem__(self, index: Literal[0, 1], value: tuple[Widget, OverlayOptions]) -> None:
         try:
             value_w, value_options = value
         except (ValueError, TypeError) as exc:
@@ -431,7 +654,7 @@ class Overlay(Widget, WidgetContainerMixin, WidgetContainerListContentsMixin):
             # normalize first, this is where errors are raised
             align_type, align_amount = normalize_align(simplify_align(align_type, align_amount), OverlayError)
             width_type, width_amount = normalize_width(simplify_width(width_type, width_amount), OverlayError)
-            valign_type, valign_amoun = normalize_valign(simplify_valign(valign_type, valign_amount), OverlayError)
+            valign_type, valign_amount = normalize_valign(simplify_valign(valign_type, valign_amount), OverlayError)
             height_type, height_amount = normalize_height(simplify_height(height_type, height_amount), OverlayError)
             self.align_type = align_type
             self.align_amount = align_amount
@@ -451,18 +674,23 @@ class Overlay(Widget, WidgetContainerMixin, WidgetContainerListContentsMixin):
             raise IndexError(f"Overlay.contents has no position {index!r}")
         self._invalidate()
 
-    def get_cursor_coords(self, size: tuple[int, int]) -> tuple[int, int] | None:
+    def get_cursor_coords(self, size: tuple[()] | tuple[int] | tuple[int, int]) -> tuple[int, int] | None:
         """Return cursor coords from top_w, if any."""
         if not hasattr(self.top_w, "get_cursor_coords"):
             return None
-        (maxcol, maxrow) = size
-        left, right, top, bottom = self.calculate_padding_filler(size, True)
+        real_size = self.pack(size, True)
+        (maxcol, maxrow) = real_size
+        left, right, top, bottom = self.calculate_padding_filler(real_size, True)
         x, y = self.top_w.get_cursor_coords((maxcol - left - right, maxrow - top - bottom))
         if y >= maxrow:  # required??
             y = maxrow - 1
         return x + left, y + top
 
-    def calculate_padding_filler(self, size: tuple[int, int], focus: bool) -> tuple[int, int, int, int]:
+    def calculate_padding_filler(
+        self,
+        size: tuple[int, int],
+        focus: bool,
+    ) -> tuple[int, int, int, int]:
         """Return (padding left, right, filler top, bottom)."""
         (maxcol, maxrow) = size
         height = None
@@ -534,7 +762,14 @@ class Overlay(Widget, WidgetContainerMixin, WidgetContainerListContentsMixin):
             )
         return left, right, top, bottom
 
-    def top_w_size(self, size, left, right, top, bottom):
+    def top_w_size(
+        self,
+        size: tuple[int, int],
+        left: int,
+        right: int,
+        top: int,
+        bottom: int,
+    ) -> tuple[()] | tuple[int] | tuple[int, int]:
         """Return the size to pass to top_w."""
         if self.width_type == WHSettings.PACK:
             # top_w is a fixed widget
@@ -545,14 +780,16 @@ class Overlay(Widget, WidgetContainerMixin, WidgetContainerListContentsMixin):
             return (maxcol - left - right,)
         return (maxcol - left - right, maxrow - top - bottom)
 
-    def render(self, size: tuple[int, int], focus: bool = False) -> CompositeCanvas:
+    def render(self, size: tuple[()] | tuple[int] | tuple[int, int], focus: bool = False) -> CompositeCanvas:
         """Render top_w overlayed on bottom_w."""
-        left, right, top, bottom = self.calculate_padding_filler(size, focus)
-        bottom_c = self.bottom_w.render(size)
+        real_size = self.pack(size, focus)
+
+        left, right, top, bottom = self.calculate_padding_filler(real_size, focus)
+        bottom_c = self.bottom_w.render(real_size)
         if not bottom_c.cols() or not bottom_c.rows():
             return CompositeCanvas(bottom_c)
 
-        top_c = self.top_w.render(self.top_w_size(size, left, right, top, bottom), focus)
+        top_c = self.top_w.render(self.top_w_size(real_size, left, right, top, bottom), focus)
         top_c = CompositeCanvas(top_c)
         if left < 0 or right < 0:
             top_c.pad_trim_left_right(min(0, left), min(0, right))
@@ -561,16 +798,31 @@ class Overlay(Widget, WidgetContainerMixin, WidgetContainerListContentsMixin):
 
         return CanvasOverlay(top_c, bottom_c, left, top)
 
-    def mouse_event(self, size: tuple[int, int], event, button: int, col: int, row: int, focus: bool) -> bool | None:
+    def mouse_event(
+        self,
+        size: tuple[()] | tuple[int] | tuple[int, int],
+        event: str,
+        button: int,
+        col: int,
+        row: int,
+        focus: bool,
+    ) -> bool | None:
         """Pass event to top_w, ignore if outside of top_w."""
         if not hasattr(self.top_w, "mouse_event"):
             return False
 
-        left, right, top, bottom = self.calculate_padding_filler(size, focus)
-        maxcol, maxrow = size
+        real_size = self.pack(size, focus)
+
+        left, right, top, bottom = self.calculate_padding_filler(real_size, focus)
+        maxcol, maxrow = real_size
         if col < left or col >= maxcol - right or row < top or row >= maxrow - bottom:
             return False
 
         return self.top_w.mouse_event(
-            self.top_w_size(size, left, right, top, bottom), event, button, col - left, row - top, focus
+            self.top_w_size(real_size, left, right, top, bottom),
+            event,
+            button,
+            col - left,
+            row - top,
+            focus,
         )
