@@ -12,28 +12,32 @@ from .constants import (
     Align,
     Sizing,
     WHSettings,
-    WrapMode,
     normalize_align,
     normalize_width,
     simplify_align,
     simplify_width,
 )
-from .widget_decoration import WidgetDecoration
+from .widget_decoration import WidgetDecoration, WidgetError, WidgetWarning
 
 if typing.TYPE_CHECKING:
     from typing_extensions import Literal
 
-    from .widget import Widget
+
+WrappedWidget = typing.TypeVar("WrappedWidget")
 
 
-class PaddingError(Exception):
-    pass
+class PaddingError(WidgetError):
+    """Padding related errors."""
 
 
-class Padding(WidgetDecoration):
+class PaddingWarning(WidgetWarning):
+    """Padding related warnings."""
+
+
+class Padding(WidgetDecoration[WrappedWidget]):
     def __init__(
         self,
-        w: Widget,
+        w: WrappedWidget,
         align: (
             Literal["left", "center", "right"] | Align | tuple[Literal["relative", WHSettings.RELATIVE], int]
         ) = Align.LEFT,
@@ -45,7 +49,7 @@ class Padding(WidgetDecoration):
         min_width: int | None = None,
         left: int = 0,
         right: int = 0,
-    ):
+    ) -> None:
         """
         :param w: a box, flow or fixed widget to pad on the left and/or right
             this widget is stored as self.original_widget
@@ -70,7 +74,7 @@ class Padding(WidgetDecoration):
 
         :param min_width: the minimum number of columns for
             self.original_widget or ``None``
-        :type min_width: int
+        :type min_width: int | None
 
         :param left: a fixed number of columns to pad on the left
         :type left: int
@@ -80,10 +84,12 @@ class Padding(WidgetDecoration):
 
         Clipping Mode: (width= ``'clip'``)
         In clipping mode this padding widget will behave as a flow
-        widget and self.original_widget will be treated as a fixed
-        widget.  self.original_widget will be clipped to fit
-        the available number of columns.  For example if align is
-        ``'left'`` then self.original_widget may be clipped on the right.
+        widget and self.original_widget will be treated as a fixed widget.
+        self.original_widget will be clipped to fit the available number of columns.
+        For example if align is ``'left'`` then self.original_widget may be clipped on the right.
+
+        Pack Mode: (width= ``'pack'``)
+        In pack mode is supported FIXED operation if it is supported by the original widget.
 
         >>> from urwid import Divider, Text, BigText, FontRegistry
         >>> from urwid.util import set_temporary_encoding
@@ -135,7 +141,7 @@ class Padding(WidgetDecoration):
 
         # convert old clipping mode width=None to width='clip'
         if width is None:
-            width = WrapMode.CLIP
+            width = WHSettings.CLIP
 
         self.left = left
         self.right = right
@@ -143,12 +149,32 @@ class Padding(WidgetDecoration):
         self._width_type, self._width_amount = normalize_width(width, PaddingError)
         self.min_width = min_width
 
-    def sizing(self):
-        if self._width_type == WrapMode.CLIP:
-            return {Sizing.FLOW}
-        return self.original_widget.sizing()
+    def sizing(self) -> frozenset[Sizing]:
+        """Widget sizing.
 
-    def _repr_attrs(self):
+        Rules:
+        * width == CLIP: only FLOW is supported, and wrapped widget should support FIXED
+        * width == GIVEN: FIXED is supported, and wrapped widget should support FLOW
+        * All other cases: use sizing of target widget
+        """
+        if self._width_type == WHSettings.CLIP:
+            return frozenset((Sizing.FLOW,))
+
+        sizing = set(self.original_widget.sizing())
+        if self._width_type == WHSettings.GIVEN:
+            if Sizing.FLOW in sizing:
+                sizing.add(Sizing.FIXED)
+
+            elif Sizing.BOX not in sizing:
+                warnings.warn(
+                    f"WHSettings.GIVEN expect BOX or FLOW widget to be used, but received {self.original_widget}",
+                    PaddingWarning,
+                    stacklevel=3,
+                )
+
+        return frozenset(sizing)
+
+    def _repr_attrs(self) -> dict[str, typing.Any]:
         attrs = dict(
             super()._repr_attrs(),
             align=self.align,
@@ -160,14 +186,18 @@ class Padding(WidgetDecoration):
         return remove_defaults(attrs, Padding.__init__)
 
     @property
-    def align(self) -> Literal["left", "center", "right"] | tuple[Literal["relative"], int]:
+    def align(
+        self,
+    ) -> Literal["left", "center", "right"] | Align | tuple[Literal["relative", WHSettings.RELATIVE], int]:
         """
         Return the padding alignment setting.
         """
         return simplify_align(self._align_type, self._align_amount)
 
     @align.setter
-    def align(self, align: Literal["left", "center", "right"] | tuple[Literal["relative"], int]) -> None:
+    def align(
+        self, align: Literal["left", "center", "right"] | Align | tuple[Literal["relative", WHSettings.RELATIVE], int]
+    ) -> None:
         """
         Set the padding alignment.
         """
@@ -193,14 +223,27 @@ class Padding(WidgetDecoration):
         self.align = align
 
     @property
-    def width(self) -> Literal["clip", "pack"] | int | tuple[Literal["relative"], int]:
+    def width(
+        self,
+    ) -> (
+        Literal["clip", "pack", WHSettings.CLIP, WHSettings.PACK]
+        | int
+        | tuple[Literal["relative", WHSettings.RELATIVE], int]
+    ):
         """
         Return the padding width.
         """
         return simplify_width(self._width_type, self._width_amount)
 
     @width.setter
-    def width(self, width: Literal["clip", "pack"] | int | tuple[Literal["relative"], int]) -> None:
+    def width(
+        self,
+        width: (
+            Literal["clip", "pack", WHSettings.CLIP, WHSettings.PACK]
+            | int
+            | tuple[Literal["relative", WHSettings.RELATIVE], int]
+        ),
+    ) -> None:
         """
         Set the padding width.
         """
@@ -225,25 +268,67 @@ class Padding(WidgetDecoration):
         )
         self.width = width
 
+    def pack(self, size: tuple[()] | tuple[int] | tuple[int, int] = (), focus: bool = False) -> tuple[int, int]:
+        if size:
+            return super().pack(size, focus)
+        if self._width_type == WHSettings.CLIP:
+            raise PaddingError("WHSettings.CLIP makes Padding FLOW-only widget")
+
+        expand = self.left + self.right
+        w_sizing = self.original_widget.sizing()
+
+        if self._width_type == WHSettings.GIVEN:
+            if Sizing.FLOW not in w_sizing:
+                warnings.warn(
+                    f"WHSettings.GIVEN expect FLOW widget to be used for FIXED pack/render, "
+                    f"but received {self.original_widget}",
+                    PaddingWarning,
+                    stacklevel=3,
+                )
+
+            return (
+                max(self._width_amount, self.min_width or 1) + expand,
+                self.original_widget.rows((self._width_amount,), focus),
+            )
+
+        if Sizing.FIXED not in w_sizing:
+            warnings.warn(
+                f"Padded widget should support FIXED sizing for FIXED render, but received {self.original_widget}",
+                PaddingWarning,
+                stacklevel=3,
+            )
+        width, height = self.original_widget.pack(size, focus)
+
+        if self._width_type == WHSettings.PACK:
+            return max(width, self.min_width or 1) + expand, height
+
+        if self._width_type == WHSettings.RELATIVE:
+            return max(int(width * 100 / self._width_amount + 0.5), self.min_width or 1) + expand, height
+
+        raise PaddingError(f"Unexpected width type: {self._width_type.upper()})")
+
     def render(
         self,
-        size: tuple[int] | tuple[int, int],
+        size: tuple[()] | tuple[int] | tuple[int, int],
         focus: bool = False,
     ) -> CompositeCanvas:
         left, right = self.padding_values(size, focus)
 
-        maxcol = size[0]
-        maxcol -= left + right
-
-        if self._width_type == WrapMode.CLIP:
+        if self._width_type == WHSettings.CLIP:
             canv = self._original_widget.render((), focus)
+        elif self._width_type == WHSettings.GIVEN:
+            canv = self._original_widget.render((self._width_amount,) + size[1:], focus)
+        elif size:
+            canv = self._original_widget.render((size[0] - (left + right),) + size[1:], focus)
         else:
-            canv = self._original_widget.render((maxcol,) + size[1:], focus)
+            canv = self._original_widget.render((), focus)
+
         if canv.cols() == 0:
             canv = SolidCanvas(" ", size[0], canv.rows())
             canv = CompositeCanvas(canv)
             canv.set_depends([self._original_widget])
             return canv
+
         canv = CompositeCanvas(canv)
         canv.set_depends([self._original_widget])
         if left != 0 or right != 0:
@@ -251,26 +336,38 @@ class Padding(WidgetDecoration):
 
         return canv
 
-    def padding_values(self, size: tuple[int] | tuple[int, int], focus: bool) -> tuple[int, int]:
+    def padding_values(
+        self,
+        size: tuple[()] | tuple[int] | tuple[int, int],
+        focus: bool,
+    ) -> tuple[int, int]:
         """Return the number of columns to pad on the left and right.
 
         Override this method to define custom padding behaviour."""
-        maxcol = size[0]
-        if self._width_type == WrapMode.CLIP:
+        if self._width_type == WHSettings.CLIP:
             width, _ignore = self._original_widget.pack((), focus=focus)
+            if not size:
+                raise PaddingError("WHSettings.CLIP makes Padding FLOW-only widget")
             return calculate_left_right_padding(
-                maxcol,
+                size[0],
                 self._align_type,
                 self._align_amount,
-                WrapMode.CLIP,
+                WHSettings.CLIP,
                 width,
                 None,
                 self.left,
                 self.right,
             )
+
         if self._width_type == WHSettings.PACK:
-            maxwidth = max(maxcol - self.left - self.right, self.min_width or 0)
-            (width, _ignore) = self._original_widget.pack((maxwidth,), focus=focus)
+            if size:
+                maxcol = size[0]
+                maxwidth = max(maxcol - self.left - self.right, self.min_width or 0)
+                (width, _ignore) = self._original_widget.pack((maxwidth,), focus=focus)
+            else:
+                (width, _ignore) = self._original_widget.pack((), focus=focus)
+                maxcol = width + self.left + self.right
+
             return calculate_left_right_padding(
                 maxcol,
                 self._align_type,
@@ -281,6 +378,18 @@ class Padding(WidgetDecoration):
                 self.left,
                 self.right,
             )
+
+        if size:
+            maxcol = size[0]
+        elif self._width_type == WHSettings.GIVEN:
+            maxcol = self._width_amount + self.left + self.right
+        else:
+            maxcol = (
+                max(self._original_widget.pack((), focus=focus)[0] * 100 // self._width_amount, self.min_width or 1)
+                + self.left
+                + self.right
+            )
+
         return calculate_left_right_padding(
             maxcol,
             self._align_type,
@@ -292,43 +401,49 @@ class Padding(WidgetDecoration):
             self.right,
         )
 
-    def rows(self, size, focus=False):
+    def rows(self, size: tuple[int], focus: bool = False) -> int:
         """Return the rows needed for self.original_widget."""
         (maxcol,) = size
         left, right = self.padding_values(size, focus)
         if self._width_type == WHSettings.PACK:
             _pcols, prows = self._original_widget.pack((maxcol - left - right,), focus)
             return prows
-        if self._width_type == WrapMode.CLIP:
+        if self._width_type == WHSettings.CLIP:
             _fcols, frows = self._original_widget.pack((), focus)
             return frows
         return self._original_widget.rows((maxcol - left - right,), focus=focus)
 
-    def keypress(self, size: tuple[int] | tuple[int, int], key: str) -> str | None:
+    def keypress(self, size: tuple[()] | tuple[int] | tuple[int, int], key: str) -> str | None:
         """Pass keypress to self._original_widget."""
-        maxcol = size[0]
         left, right = self.padding_values(size, True)
-        maxvals = (maxcol - left - right,) + size[1:]
-        return self._original_widget.keypress(maxvals, key)
+        if size:
+            maxvals = (size[0] - left - right,) + size[1:]
+            return self._original_widget.keypress(maxvals, key)
+        return self._original_widget.keypress((), key)
 
-    def get_cursor_coords(self, size: tuple[int] | tuple[int, int]) -> tuple[int, int] | None:
+    def get_cursor_coords(self, size: tuple[()] | tuple[int] | tuple[int, int]) -> tuple[int, int] | None:
         """Return the (x,y) coordinates of cursor within self._original_widget."""
         if not hasattr(self._original_widget, "get_cursor_coords"):
             return None
+
         left, right = self.padding_values(size, True)
-        maxcol = size[0]
-        maxvals = (maxcol - left - right,) + size[1:]
-        if maxvals[0] == 0:
-            return None
+        if size:
+            maxvals = (size[0] - left - right,) + size[1:]
+            if maxvals[0] == 0:
+                return None
+        else:
+            maxvals = ()
+
         coords = self._original_widget.get_cursor_coords(maxvals)
         if coords is None:
             return None
+
         x, y = coords
         return x + left, y
 
     def move_cursor_to_coords(
         self,
-        size: tuple[int] | tuple[int, int],
+        size: tuple[()] | tuple[int] | tuple[int, int],
         x: int,
         y: int,
     ) -> bool:
@@ -338,45 +453,59 @@ class Padding(WidgetDecoration):
         """
         if not hasattr(self._original_widget, "move_cursor_to_coords"):
             return True
+
         left, right = self.padding_values(size, True)
-        maxcol = size[0]
-        maxvals = (maxcol - left - right,) + size[1:]
+        if size:
+            maxcol = size[0]
+            maxvals = (maxcol - left - right,) + size[1:]
+        else:
+            maxcol = self.pack((), True)[0]
+            maxvals = ()
+
         if isinstance(x, int):
             if x < left:
                 x = left
             elif x >= maxcol - right:
                 x = maxcol - right - 1
             x -= left
+
         return self._original_widget.move_cursor_to_coords(maxvals, x, y)
 
     def mouse_event(
         self,
-        size: tuple[int] | tuple[int, int],
-        event,
+        size: tuple[()] | tuple[int] | tuple[int, int],
+        event: str,
         button: int,
         x: int,
         y: int,
         focus: bool,
-    ):
+    ) -> bool | None:
         """Send mouse event if position is within self._original_widget."""
         if not hasattr(self._original_widget, "mouse_event"):
             return False
 
         left, right = self.padding_values(size, focus)
-        maxcol = size[0]
-        if x < left or x >= maxcol - right:
-            return False
-        maxvals = (maxcol - left - right,) + size[1:]
+        if size:
+            maxcol = size[0]
+            if x < left or x >= maxcol - right:
+                return False
+            maxvals = (maxcol - left - right,) + size[1:]
+        else:
+            maxvals = ()
+
         return self._original_widget.mouse_event(maxvals, event, button, x - left, y, focus)
 
-    def get_pref_col(self, size: tuple[int] | tuple[int, int]) -> int | None:
+    def get_pref_col(self, size: tuple[()] | tuple[int] | tuple[int, int]) -> int | None:
         """Return the preferred column from self._original_widget, or None."""
         if not hasattr(self._original_widget, "get_pref_col"):
             return None
 
         left, right = self.padding_values(size, True)
-        maxcol = size[0]
-        maxvals = (maxcol - left - right,) + size[1:]
+        if size:
+            maxvals = (size[0] - left - right,) + size[1:]
+        else:
+            maxvals = ()
+
         x = self._original_widget.get_pref_col(maxvals)
         if isinstance(x, int):
             return x + left
@@ -387,7 +516,7 @@ def calculate_left_right_padding(
     maxcol: int,
     align_type: Literal["left", "center", "right"] | Align,
     align_amount: int,
-    width_type: Literal["fixed", "relative", "clip"],
+    width_type: Literal["fixed", "relative", "clip", "given", WHSettings.RELATIVE, WHSettings.CLIP, WHSettings.GIVEN],
     width_amount: int,
     min_width: int | None,
     left: int,
@@ -430,7 +559,7 @@ def calculate_left_right_padding(
     """
     if width_type == WHSettings.RELATIVE:
         maxwidth = max(maxcol - left - right, 0)
-        width = int_scale(width_amount, 101, maxwidth + 1)
+        width = int(maxwidth * width_amount / 100 + 0.5)
         if min_width is not None:
             width = max(width, min_width)
     else:
@@ -455,7 +584,7 @@ def calculate_left_right_padding(
         left += shift
 
     # only clip if width_type == 'clip'
-    if width_type != WrapMode.CLIP and (left < 0 or right < 0):
+    if width_type != WHSettings.CLIP and (left < 0 or right < 0):
         left = max(left, 0)
         right = max(right, 0)
 
